//! Acme client that supports multiple CAs and configs for sites that can be seperate from the mainconfig
#![allow(dead_code)]
#![allow(clippy::clone_on_copy)]
#![allow(clippy::identity_op)]

pub(crate) mod consts;
pub(crate) mod macros;
pub(crate) mod prelude;
pub(crate) mod process;
pub(crate) mod types;
pub(crate) mod utils;

use crate::{
    consts::*,
    prelude::*,
    types::{
        config::{
            GeneralConfig,
            SiteConfig,
        },
        structs::{
            Arguments,
            ProcessorArgs,
        },
    },
};
use acme2_eab::Directory;
use async_scoped::TokioScope;
use clap::Parser;
use env_logger::init as log_init;
use libsystemd::daemon;
use log::*;
use openssl::{
    self,
    pkey::{
        PKey,
        Private,
    },
};
use reqwest::tls::Version;
use std::{
    collections::{
        HashMap,
        HashSet,
    },
    path::{
        Path,
        PathBuf,
    },
    process::exit,
    str::FromStr,
    sync::Arc,
    time::Duration,
};
use tokio::{
    fs::{
        create_dir_all,
        read_dir,
    },
    io::AsyncReadExt,
    sync::Mutex,
};
use tokio_stream::{
    StreamExt,
    wrappers::ReadDirStream,
};
use types::structs::Error;


fn default_client() -> Result<reqwest::Client, Error> {
    reqwest::Client::builder()
        .min_tls_version(Version::TLS_1_2)
        .https_only(true)
        .pool_max_idle_per_host(POOL_SIZE)
        .build()
        .map_err(Error::from_display)
}

async fn load_privkey(path: PathBuf) -> Result<PKey<Private>, Error> {
    let mut file = match FILE_MODE.open(path).await {
        Ok(file) => file,
        Err(error) => return Error::err(format!("Failed to open Private Key: {error}")),
    };
    let mut data = String::new();
    if let Err(error) = file.read_to_string(&mut data).await {
        return Error::err(format!("Failed to read data for the key: {error}"));
    }
    match PKey::private_key_from_pem(data.as_bytes()) {
        Ok(key) => Ok(key),
        Err(error) => Error::err(format!("Failed to parse pem data: {error}")),
    }
}

async fn racme(flags: Arguments) -> Result<(), Error> {
    let client = default_client()?;
    let systemd_access = daemon::booted();
    let mainconfig = {
        let file = match FILE_MODE.open(flags.config).await {
            Ok(file) => file,
            Err(error) => {
                return Error::err(format!("error reading the config: {error}"));
            },
        };
        GeneralConfig::from_file(file).await
    };
    trace!("Parsed Config: {mainconfig:?}");
    let files = {
        let rd = match read_dir(mainconfig.sites_path.clone()).await {
            Ok(rd) => rd,
            Err(error) => {
                return Error::err(format!("could not read files from sites dir: {error}"));
            },
        };
        ReadDirStream::new(rd)
    };
    let mut siteconfigs = Vec::new();
    for filename in files.filter(Result::is_ok).map(|file| file.unwrap().path()).collect::<Vec<PathBuf>>().await {
        let file = match FILE_MODE.open(filename.clone()).await {
            Ok(file) => file,
            Err(error) => {
                warn!("Failed to read the configfile {}: {}", filename.display(), error);
                continue;
            },
        };
        let mut site = SiteConfig::from_file(file).await;
        site.name = filename.file_stem().unwrap().to_string_lossy().to_string();
        siteconfigs.push(site);
    }
    let used = siteconfigs.iter().map(|s| s.ca.clone()).collect::<HashSet<_>>();
    debug!("Used CAs: {used:?}");

    let mut directories = HashMap::<String, Arc<Directory>>::new();

    let mut accounts = HashMap::new();
    let accountpath = Path::new(&mainconfig.accounts_path).to_path_buf();
    if let Err(error) = create_dir_all(accountpath.clone()).await {
        return Error::err(format!("Failed to create the directory for the accounts: {error}"));
    }
    for (name, ca) in mainconfig.ca.iter().filter(|(name, _)| used.contains(name.to_owned())) {
        process::accounts(name, ca, &mut directories, &mut accounts, &client, accountpath.clone()).await;
    }

    let restart_services = Mutex::new(HashSet::<String>::new());
    let reload_services = Mutex::new(HashSet::<String>::new());
    let certs = Path::new(&mainconfig.certificates_path).to_path_buf();

    if !certs.exists() {
        if let Err(error) = create_dir_all(certs.clone()).await {
            return Error::err(format!("Failed to create directory for all the certificates: {error}"));
        }
    }
    let challengepath = mainconfig.http_challenge_path.and_then(|path| PathBuf::from_str(path.as_str()).ok());
    let dnsserver = None;

    unsafe {
        TokioScope::scope_and_collect(|scope| {
            for site in siteconfigs {
                if let Some(account) = accounts.get(&site.ca) {
                    scope.spawn(process::site(ProcessorArgs::new(
                        site,
                        Arc::clone(account),
                        &reload_services,
                        &restart_services,
                        certs.clone(),
                        challengepath.clone(),
                        dnsserver.clone(),
                    )));
                } else {
                    error!("Could not process site {} because of previous errors", site.name)
                }
            }
        })
    }
    .await;

    if systemd_access {
        process::services(restart_services.into_inner(), reload_services.into_inner()).await;
    }
    Ok(())
}

fn main() {
    log_init();
    let runtime = match tokio::runtime::Builder::new_current_thread().enable_all().build() {
        Ok(runtime) => runtime,
        Err(error) => {
            error!("Could not initialize Tokio runtime: {error}");
            exit(2)
        },
    };
    let result = runtime.block_on(racme(Arguments::parse()));
    runtime.shutdown_timeout(Duration::from_secs(1));
    if let Err(error) = result {
        error!("{error}");
        exit(1)
    }
}
